# -*- coding: utf-8 -*-
# @Time    : 2023/8/21 上午12:06
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
# 负责任务前后的APIKey信息管理和验证机制(认证可信平台) 平台:ID
import time
from enum import Enum
from typing import List, Union, Optional

from pydantic import field_validator, ConfigDict, BaseModel, Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from ...sdk.endpoint import Driver


class UserDriverMode(Enum):
    public = 100
    """公共环境变量，只有预先设定的规则管控了Driver的配置，此情况无论如何都不能配置其他endpoint"""

    private = 200
    """私有 环境变量/私有端点，要求用户无论如何自己配置，用户可以自己产生 Driver"""

    proxy_public = 300
    """代理公共环境变量，也就是额外的token计费系统控制了公共环境变量的使用"""


# 基本单元

class UserCost(BaseModel):
    """用户消费记录
    """

    class Cost(BaseModel):
        """消费记录细节
        """
        cost_by: str = Field("chat", description="环节")
        token_usage: int = Field(0)
        token_uuid: Optional[str] = Field(None, description="Api Key 的 hash")
        llm_model: Optional[str] = Field(None, description="Model Name")
        provide_type: int = Field(None, description="认证模式")

        @classmethod
        def by_function(cls, function_name: str,
                        token_usage: int,
                        token_uuid: str,
                        llm_model: str,
                        ):
            return cls(cost_by=function_name, token_usage=token_usage, token_uuid=token_uuid, model_name=llm_model)

    request_id: str = Field(default=None, description="请求 UUID")
    uid: str = Field(default=None, description="用户 UID ,注意是平台+用户")
    cost: Cost = Field(default=None, description="消费记录")
    cost_time: int = Field(default=None, description="消费时间")
    meta: dict = Field(default={}, description="元数据")

    @classmethod
    def create_from_function(
            cls,
            uid: str,
            request_id: str,
            cost_by: str,
            token_usage: int,
            token_uuid: str,
            model_name: str,
    ):
        return cls(
            request_id=request_id,
            uid=uid,
            cost=cls.Cost.by_function(
                function_name=cost_by,
                token_usage=token_usage,
                token_uuid=token_uuid,
                llm_model=model_name,
            ),
            cost_time=int(time.time()),
        )

    @classmethod
    def create_from_task(
            cls,
            uid: str,
            request_id: str,
            cost: Cost,
    ):
        return cls(
            request_id=request_id,
            uid=uid,
            cost=cost,
            cost_time=int(time.time()),
        )

    model_config = ConfigDict(extra="ignore",
                              arbitrary_types_allowed=True,
                              validate_assignment=True,
                              validate_default=True
                              )


class UserConfig(BaseSettings):
    """
    :tip 注意此类和公共环境变量的区别！禁止用户使用 公共变量 请求 私有端点！
    """

    class LlmConfig(BaseModel):
        """用户配置
        driver 作为一个单例模式
        其他 `公共授权` 组件！
        """
        driver: Optional[Driver] = Field(None, description="私有端点配置")
        token: Optional[str] = Field(None, description="代理认证系统的token")
        provider: Optional[str] = Field(None, description="认证平台")

        @property
        def mode(self):
            """
            :return: 返回模式
            """
            if self.driver:
                if self.driver.api_key:
                    return UserDriverMode.private
            if self.token and not self.driver:
                return UserDriverMode.proxy_public
            return UserDriverMode.public

        @classmethod
        def default(cls):
            return cls()

        def set_proxy_public(self, token: str, provider: str):
            self.provider = provider
            self.token = token
            return self

        @field_validator("provider")
        def upper_provider(cls, v):
            if v:
                return v.upper()
            return v

    class PluginConfig(BaseModel):
        block_list: List[str] = Field([], description="黑名单")

        @classmethod
        def default(cls):
            return cls()

        def block(self, plugin_name: str) -> "UserConfig.PluginConfig":
            if plugin_name not in self.block_list:
                self.block_list.append(plugin_name)
            return self

        def unblock(self, plugin_name: str) -> "UserConfig.PluginConfig":
            if plugin_name in self.block_list:
                self.block_list.remove(plugin_name)
            return self

    created_time: int = Field(default=int(time.time()), description="创建时间")
    last_use_time: int = Field(default=int(time.time()), description="最后使用时间")
    uid: Union[str, int] = Field(None, description="用户UID")
    plugin_subs: PluginConfig = Field(default_factory=PluginConfig.default, description="插件订阅")
    llm_driver: LlmConfig = Field(default_factory=LlmConfig.default, description="驱动")

    @field_validator("uid")
    def check_user_id(cls, v):
        if v:
            return str(v)
        return v

    model_config = SettingsConfigDict(extra="ignore",
                                      frozen=True,
                                      arbitrary_types_allowed=True,
                                      validate_assignment=True,
                                      validate_default=True
                                      )



# -*- coding: utf-8 -*-
from pydantic import ConfigDict

__package__name__ = "llmkira.extra.plugins.search"
__plugin_name__ = "search_in_google"
__openapi_version__ = "20231111"

from llmkira.extra.user import CostControl, UserCost
from llmkira.middleware.llm_provider import GetAuthDriver
from llmkira.sdk import resign_plugin_executor
from llmkira.sdk.endpoint import openai
from llmkira.sdk.func_calling import verify_openapi_version

verify_openapi_version(__package__name__, __openapi_version__)
from loguru import logger  # noqa: E402
from pydantic import BaseModel  # noqa: E402

from llmkira.schema import RawMessage  # noqa: E402
from llmkira.sdk.func_calling import BaseTool, PluginMetadata  # noqa: E402
from llmkira.sdk.func_calling.schema import FuncPair  # noqa: E402
from llmkira.sdk.schema import create_short_task, Function  # noqa: E402
from llmkira.task import Task, TaskHeader  # noqa: E402
from typing import TYPE_CHECKING  # noqa: E402

if TYPE_CHECKING:
    from llmkira.sdk.schema import TaskBatch
search = Function(
    name=__plugin_name__,
    description="Search/validate on google.com.[ONLY IF NECESSARY]",
).update_config(
    config=Function.FunctionExtra(
        system_prompt="Search only if necessary",
    )
)
search.add_property(
    property_name="keywords",
    property_description="question entered in the search website",
    property_type="string",
    required=True,
)


@resign_plugin_executor(function=search)
def search_on_duckduckgo(search_sentence: str, key_words: str = None):
    logger.debug(f"Plugin --search_on_duckduckgo {search_sentence}")
    from duckduckgo_search import DDGS

    # 内存优化抛弃 NLP
    # from llmkira.sdk.filter import Sublimate
    sort_text = []
    link_refer = {}
    with DDGS(timeout=20) as ddgs:
        search_result = ddgs.text(search_sentence)
        for r in search_result:
            _title = r.get("title")
            _href = r.get("href")
            _body = r.get("body")
            link_refer[_body] = _href
            sort_text.append((_body, _title, _href))
    # must_key = [key_words] if key_words else None
    sorted_result = sort_text
    # sorted_result = Sublimate(sort_text).valuation(match_sentence=search_sentence, match_keywords=must_key)
    valuable_result = [item[0] for item in sorted_result[:4]]
    # 构建单条内容
    clues = []
    for key, item in enumerate(valuable_result):
        clues.append(
            f"\nPage #{key}\n🔍Contents:{item}\n"
            f"🔗Source:{link_refer.get(item, 'https://google.com/')}\n"
        )
    content = "\n".join(clues)
    return (
        "[🔍SearchPage]\n"
        + content
        + (
            "\n[Page End]"
            "\n[ReplyFormat:`$summary_answer \n [$index]($source_link) * num` to mark links]"
        )
    )


class Search(BaseModel):
    keywords: str
    model_config = ConfigDict(extra="allow")


class SearchTool(BaseTool):
    """
    搜索工具
    """

    silent: bool = False
    function: Function = search
    keywords: list = [
        "怎么",
        "How",
        "件事",
        "牢大",
        "作用",
        "知道",
        "什么",
        "认识",
        "What",
        "http",
        "what",
        "who",
        "how",
        "Who",
        "Why",
        "作品",
        "why",
        "Where",
        "了解",
        "简述",
        "How to",
        "是谁",
        "how to",
        "解释",
        "怎样的",
        "新闻",
        "ニュース",
        "电影",
        "番剧",
        "アニメ",
        "2022",
        "2023",
        "请教",
        "介绍",
        "怎样",
        "吗",
        "么",
        "？",
        "?",
        "呢",
        "评价",
        "搜索",
        "百度",
        "谷歌",
        "bing",
        "谁是",
        "上网",
    ]

    def pre_check(self):
        try:
            from duckduckgo_search import DDGS  # noqa: F401

            # from llmkira.sdk.filter import Sublimate
            return True
        except ImportError as e:
            logger.warning(
                f"plugin:package <duckduckgo_search> not found,please install it first:{e}"
            )
            return False

    def func_message(self, message_text, **kwargs):
        """
        如果合格则返回message，否则返回None，表示不处理
        """
        for i in self.keywords:
            if i in message_text:
                return self.function
        # 正则匹配
        if self.pattern:
            match = self.pattern.match(message_text)
            if match:
                return self.function
        return None

    async def failed(
        self,
        task: "TaskHeader",
        receiver: "TaskHeader.Location",
        exception,
        env: dict,
        arg: dict,
        pending_task: "TaskBatch",
        refer_llm_result: dict = None,
        **kwargs,
    ):
        _meta = task.task_meta.reply_notify(
            plugin_name=__plugin_name__,
            callback=[
                TaskHeader.Meta.Callback.create(
                    name=__plugin_name__,
                    function_response=f"Run Failed {exception}",
                    tool_call_id=pending_task.get_batch_id(),
                )
            ],
            write_back=True,
            release_chain=True,
        )
        await Task(queue=receiver.platform).send_task(
            task=TaskHeader(
                sender=task.sender,
                receiver=receiver,
                task_meta=_meta,
                message=[
                    RawMessage(
                        user_id=receiver.user_id,
                        chat_id=receiver.chat_id,
                        text=f"🍖{__plugin_name__} Run Failed：{exception}",
                    )
                ],
            )
        )

    @staticmethod
    async def llm_task(plugin_name, task: TaskHeader, task_desc: str, raw_data: str):
        assert task_desc == raw_data, "STOP USE THE FUNCTION"
        # TODO: 转换为通用
        logger.info("llm_tool:{}".format(task_desc))
        auth_client = GetAuthDriver(uid=task.sender.uid)
        driver = await auth_client.get()
        endpoint = openai.Openai.init(
            driver=driver,
            temperature=0.1,
            messages=create_short_task(task_desc=task_desc, refer=raw_data),
        )
        # 调用Openai
        result = await endpoint.create()
        # 记录消耗
        await CostControl.add_cost(
            cost=UserCost.create_from_function(
                uid=task.sender.uid,
                request_id=result.id,
                cost_by=plugin_name,
                token_usage=result.usage.total_tokens,
                token_uuid=driver.uuid,
                model_name=driver.model,
            )
        )
        assert result.default_message.content, "llm_task.py:llm_task:content is None"
        return result.default_message.content

    async def callback(
        self,
        task: "TaskHeader",
        receiver: "TaskHeader.Location",
        env: dict,
        arg: dict,
        pending_task: "TaskBatch",
        refer_llm_result: dict = None,
        **kwargs,
    ):
        return True

    async def run(
        self,
        task: "TaskHeader",
        receiver: "TaskHeader.Location",
        arg: dict,
        env: dict,
        pending_task: "TaskBatch",
        refer_llm_result: dict = None,
    ):
        """
        处理message，返回message
        """

        _set = Search.model_validate(arg)
        _search_result = search_on_duckduckgo(_set.keywords)
        # META
        _meta = task.task_meta.reply_raw(
            plugin_name=__plugin_name__,
            callback=[
                TaskHeader.Meta.Callback.create(
                    name=__plugin_name__,
                    function_response=str(_search_result),
                    tool_call_id=pending_task.get_batch_id(),
                )
            ],
        )
        await Task(queue=receiver.platform).send_task(
            task=TaskHeader(
                sender=task.sender,  # 继承发送者
                receiver=receiver,  # 因为可能有转发，所以可以单配
                task_meta=_meta,
                message=[
                    RawMessage(
                        user_id=receiver.user_id,
                        chat_id=receiver.chat_id,
                        text="🔍 Searching Done",
                    )
                ],
            )
        )


__plugin_meta__ = PluginMetadata(
    name=__plugin_name__,
    description="Search fact on google.com",
    usage="以问号结尾的句子即可触发",
    openapi_version=__openapi_version__,
    function={FuncPair(function=search, tool=SearchTool)},
)



# -*- coding: utf-8 -*-
# @Time    : 2023/11/13 下午3:55
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
from typing import Optional, List

from pydantic import BaseModel, Field, model_validator


class PluginExternal(BaseModel):
    class Install(BaseModel):
        shell: Optional[str] = Field(None, title="shell安装命令")
        pypi: Optional[str] = Field(None, title="pypi安装命令")
        github: Optional[str] = Field(None, title="github安装命令")

        @model_validator(mode="after")
        def available_check(self):
            if not any([self.shell, self.pypi, self.github]):
                raise ValueError("At least one of install must be available")
            return self

    plugin_name: str = Field(..., title="插件名称")
    plugin_link: str = Field(..., title="插件链接")
    plugin_desc: str = Field(..., title="插件描述")
    plugin_functions: List[str] = Field(..., title="插件功能")
    org_id: Optional[str] = Field(None, title="组织ID")
    author_id: str = Field(..., title="作者ID")
    plugin_install: Install = Field(..., title="插件安装方式")


class PluginTestResultExport(BaseModel):
    plugin: PluginExternal = Field(..., title="插件信息")
    test_pass: bool = Field(..., title="测试是否通过")
    out_date: bool = Field(..., title="是否过期")



# -*- coding: utf-8 -*-
# @Time    : 2023/10/21 下午2:29
# @Author  : sudoskys
# @File    : event.py
# @Software: PyCharm

from pydantic import ConfigDict, BaseModel


def help_message():
    return """
*Command*

!chat - chat with me in a serious way :(
@<myname> - Chat with me in a simple way :)
!task - chat with function_enable
!ask - chat with function_disable
!auth - Auth a task

*Slash Command*

`/help` - Help message
`/tool` - Tool list
`/clear` - forget...you
`/auth` - activate a task (my power),but outside the thread
`/bind` - RssUrl.....
`/unbind` - RssUrl.....
`/set_endpoint` - <apikey>#<endpoint> for your own backend
`/clear_endpoint` - clear endpoint and key
`/env` - set environment variable
`/token` - bind your service provider token
`/token_clear` - clear your service provider token
`/func_ban` - ban function
`/func_unban` - unban function

Make sure you invite me before you call me in channel, wink~

*DONT SHARE YOUR TOKEN/ENDPOINT PUBLIC!*

- Please confirm that that bot instance is secure, some plugins may be dangerous on unsafe instance.
"""


class SlashCommand(BaseModel):
    """
    https://api.slack.com/interactivity/slash-commands#app_command_handling
    token=gIkuvaNzQIHg97ATvDxqgjtO
    &team_id=T0001
    &team_domain=example
    &enterprise_id=E0001
    &enterprise_name=Globular%20Construct%20Inc
    &channel_id=C2147483705
    &channel_name=test
    &user_id=U2147483697
    &user_name=Steve
    &command=/weather
    &text=94070
    &response_url=https://hooks.slack.com/commands/1234/5678
    &trigger_id=13345224609.738474920.8088930838d88f008e0
    &api_app_id=A123456
    """
    token: str = None
    team_id: str = None
    team_domain: str = None
    enterprise_id: str = None
    enterprise_name: str = None
    channel_id: str = None
    channel_name: str = None
    user_id: str = None
    user_name: str = None
    command: str = None
    text: str = None
    response_url: str = None
    trigger_id: str = None
    api_app_id: str = None
    model_config = ConfigDict(extra="allow")


class SlackChannelInfo(BaseModel):
    id: str = None
    name: str = None
    is_channel: bool = None
    is_group: bool = None
    is_im: bool = None
    is_mpim: bool = None
    is_private: bool = None
    created: int = None
    is_archived: bool = None
    is_general: bool = None
    unlinked: int = None
    name_normalized: str = None
    is_shared: bool = None
    is_org_shared: bool = None
    is_pending_ext_shared: bool = None
    pending_shared: list = None
    context_team_id: str = None
    updated: int = None
    parent_conversation: str = None
    creator: str = None
    is_ext_shared: bool = None
    shared_team_ids: list = None
    pending_connected_team_ids: list = None
    is_member: bool = None
    last_read: str = None
    topic: dict = None
    purpose: dict = None
    previous_names: list = None
    model_config = ConfigDict(extra="allow")
